/*
 *                            *
 *                           * *
 *                          *   *
 *                     ***************
 *                      * *       * *
 *                       *  MUMPS  *
 *                      * *       * *
 *                     ***************
 *                          *   *
 *                           * *
 *                            *
 *
 *   merge.c
 *    Implementation of FreeM MERGE command
 *
 *  
 *   Author: John P. Willis <jpw@coherent-logic.com>
 *    Copyright (C) 1998 MUG Deutschland
 *    Copyright (C) 2020 Coherent Logic Development LLC
 *
 *
 *   This file is part of FreeM.
 *
 *   FreeM is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Affero Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   FreeM is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Affero Public License for more details.
 *
 *   You should have received a copy of the GNU Affero Public License
 *   along with FreeM.  If not, see <https://www.gnu.org/licenses/>.
 *
 **/

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "mpsdef.h"
#include "merr.h"

void merge(char *dst, char *src)
{

    void (*reader)(short, char *, char *);
    void (*writer)(short, char *, char *);

    char *dbuf; /* dest. key buffer */
    char *sbuf; /* src. key buffer */

    char *buf;  /* scratch space */
    char *kbuf; /* key buffer */
    char *tmp;
    char *next_key;

    short src_has_subs;
    short dst_has_subs;

    char *src_vn;
    char *src_subs;

    char *dst_vn;
    char *dst_subs;

    char *fw_key;

    short d;    /* $DATA result */

    if ((buf = malloc (STRLEN + 1)) == NULL) {
        ierr = MEMOV;
        return;
    }

    if ((kbuf = malloc (STRLEN + 1)) == NULL) {
        ierr = MEMOV;
        return;
    }

    if ((tmp = malloc (STRLEN + 1)) == NULL) {
        ierr = MEMOV;
        return;
    }

    if ((next_key = malloc (STRLEN + 1)) == NULL) {
        ierr = MEMOV;
        return;
    }

    if ((dbuf = malloc (STRLEN + 1)) == NULL) {
        ierr = MEMOV;
        return;
    }

    if ((sbuf = malloc (STRLEN + 1)) == NULL) {
        ierr = MEMOV;
        return;
    }

    if ((src_vn = malloc (STRLEN + 1)) == NULL) {
        ierr = MEMOV;
        return;
    }

    if ((src_subs = malloc (STRLEN + 1)) == NULL) {
        ierr = MEMOV;
        return;
    }

    if ((dst_vn = malloc (STRLEN + 1)) == NULL) {
        ierr = MEMOV;
        return;
    }

    if ((dst_subs = malloc (STRLEN + 1)) == NULL) {
        ierr = MEMOV;
        return;
    }

    if ((fw_key = malloc (STRLEN + 1)) == NULL) {
        ierr = MEMOV;
        return;
    }

    /* isolate the source variable name and subscripts */
    src_subs[0] = NUL;
    if (strchr (src, DELIM) != NULL) {
        src_vn = strtok (src, "\202");
        src_has_subs = TRUE;

        src_subs = strchr (src, DELIM) + 1;
    }
    else {
        src_vn = strtok (src, "\201");
        src_has_subs = FALSE;
    }

    /* isolate the destination variable name and subscripts */
    dst_subs[0] = NUL;
    if (strchr (dst, DELIM) != NULL) {
        dst_vn = strtok (dst, "\202");
        dst_has_subs = TRUE;

        dst_subs = strchr (dst, DELIM) + 1;
    }
    else {
        dst_vn = strtok (dst, "\201");
        dst_has_subs = FALSE;
    }

    set_io(UNIX);
    printf("src_vn = '%s' src_subs = '%s'\n", src_vn, src_subs);
    printf("dst_vn = '%s' dst_subs = '%s'\n", dst_vn, dst_subs);
    set_io(MUMPS);

    /* find out the nature of src and dst (local, gbl, or ssvn)
       and set up the reader and writer function pointers. */
    key_to_name (dbuf, dst, stlen (dst));
    key_to_name (sbuf, src, stlen (src));

    if (dbuf[0] != '^') {
        writer = &symtab;
    }
    else if (dbuf[1] != '$') {
        writer = &global; 
    }
    else {
        writer = &ssvn;
    }

    if (sbuf[0] != '^') {
        
        set_io (UNIX);
        fprintf (stderr, "MERGE from local variables not yet supported. C'est la FreeM...\n");
        set_io (MUMPS);

        ierr = OK;
        return;

        reader = &symtab;
    }
    else if (sbuf[1] != '$') {
        reader = &global; 
    }
    else {
        reader = &ssvn;
    }

    /* TODO: throw an error if src is descended from dst or vice-versa */


    stcpy (next_key, src);
    stcnv_c2m (next_key);

    while(1) {

        /* find out whether src has data, children, or both */
        (*reader)(dat, next_key, buf);

        stcnv_m2c (buf);
        d = atoi (buf);

        
        switch (d) {

            case 0:     /* undefined */
                ierr = UNDEF;
                return;

            case 1:     /* data only */
            case 10:    /* children only */
            case 11:    /* data and children */  

                /* munge together the key that will be used for any writes */

                if (strchr (next_key, DELIM)) {

                    /* next_key has subscripts */
                    
                    tmp = strtok (next_key, "\202");
                    snprintf (fw_key, STRLEN - 1, "%s%s\202%s\201", dst_vn, dst_subs, tmp);

                }
                else {
                    
                    snprintf (fw_key, STRLEN - 1, "%s\201", dst_vn);
                
                }

                set_io(UNIX);
                printf("next_key = '%s' fw_key = '%s' buf= '%s'\n", next_key, fw_key, buf);
                set_io(MUMPS);

                (*reader)(get_sym, next_key, buf);
                (*writer)(set_sym, fw_key, buf);
                break;

        }

        (*reader)(query, next_key, buf);

        if (buf[0] == EOL) break;

        name_to_key (next_key, buf, stlen (buf));

    }
    
    ierr = OK;
    return;

}
