/*
 *                            *
 *                           * *
 *                          *   *
 *                     ***************
 *                      * *       * *
 *                       *  MUMPS  *
 *                      * *       * *
 *                     ***************
 *                          *   *
 *                           * *
 *                            *
 *
 *   ri.c
 *    FreeM routine import utility
 *
 *  
 *   Author: John P. Willis <jpw@coherent-logic.com>
 *    Copyright (C) 1998 MUG Deutschland
 *    Copyright (C) 2020 Coherent Logic Development LLC
 *
 *
 *   This file is part of FreeM.
 *
 *   FreeM is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Affero Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   FreeM is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Affero Public License for more details.
 *
 *   You should have received a copy of the GNU Affero Public License
 *   along with FreeM.  If not, see <https://www.gnu.org/licenses/>.
 *
 **/

#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>

#ifdef __linux__
 #include <linux/limits.h>
#endif

#if !defined(PATH_MAX) && defined(__sun__)
# include <limits.h>
#endif

#if !defined(PATH_MAX) && defined(EMSCRIPTEN)
# define PATH_MAX 1024
#endif

#if !defined(PATH_MAX) && defined(__APPLE__)
# define PATH_MAX 4096
#endif

#if !defined(PATH_MAX) && defined(_SCO_DS)
# define PATH_MAX 4096
#endif

#if !defined(PATH_MAX) && defined(__gnu_hurd__)
# define PATH_MAX 1024
#endif

#if defined(__NetBSD__) || defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__AMIGA)
# include <sys/syslimits.h>
#endif

#if !defined(_SCO_DS) && !defined(__AMIGA) && !defined(_AIX)
# include <getopt.h>
#endif

#define FM_STR_MAX 258

#define FALSE   0
#define TRUE    1

char config_file[4096];

#include "iniconf.h"


int main(int argc, char **argv)
{
    
    int usr_loaded = 0;
    int pct_loaded = 0;
    int next_routine_flag = FALSE;
    int rtn_open = FALSE;
    int rtn_overwrite = FALSE;
    long ln = 0;

    char pct_rtn_path[PATH_MAX];
    char usr_rtn_path[PATH_MAX];
    char namespace[FM_STR_MAX];
    char roufile[PATH_MAX];
    char line[FM_STR_MAX];  
    
    char *filename = NULL;
    FILE *archive = NULL;
    FILE *rtn = NULL;

    namespace[0] = '\0';

#if !defined(_SCO_DS) && !defined(__AMIGA) && !defined(_AIX)
    char c;
    int option_index = 0;

    struct option long_options[] = {
        {"namespace", required_argument, 0, 'n'},
        {"file", required_argument, 0, 'f'},
        {"overwrite", no_argument, &rtn_overwrite, TRUE}
    };
#endif
    
    if((filename = (char *) malloc(PATH_MAX)) == NULL) {
        fprintf(stderr, "Memory allocation failure.\n");

        return 1;
    }

    filename[0] = '\0';

#if !defined(_SCO_DS) && !defined(__AMIGA) && !defined(_AIX)

    while(1) {
        c = getopt_long(argc, argv, "n:f:o", long_options, &option_index);

        if (c == -1) break;

        switch(c) {

            case 'n': /* namespace */
                strncpy(namespace, optarg, FM_STR_MAX);               
                break;

            case 'f':
                strncpy(filename, optarg, PATH_MAX);
                break;

        } 

    }

#else
    
    {

        char c;
        extern char *optarg;
        extern int optind, optopt;

        while ((c = getopt(argc, argv, "n:f:o")) != -1) {
            
            switch(c) {

                case 'n': /* namespace */
                    strncpy(namespace, optarg, FM_STR_MAX);               
                    break;

                case 'f':
                    strncpy(filename, optarg, PATH_MAX);
                    break;

            }    
        
        }

    }

#endif  

    if((!strlen(namespace)) || (!strlen(filename))) {
        fprintf(stderr, "Must specify --namespace and --file.\n");

        return 2;
    }

    if(get_conf("SYSTEM", "routines_path", pct_rtn_path) == FALSE) {
        fprintf(stderr, "Could not determine percent routine access path from configuration for namespace '%s'.\n", namespace);

        return 3;
    }

    if(get_conf(namespace, "routines_path", usr_rtn_path) == FALSE) {
        fprintf(stderr, "Could not determine user routine access path from configuration for namespace '%s'.\n", namespace);

        return 4;
    }

    if((archive = fopen(filename, "r")) == NULL) {
        fprintf(stderr, "Could not open routine archive '%s'.\n", filename);

        return 5;
    }

    printf("\n\nFreeM Routine Import from '%s'\n\n", filename);

    printf("  * Using FreeM namespace %s\n", namespace);
    printf("  * Percent routines will be loaded into %s\n", pct_rtn_path);
    printf("  * User routines will be loaded into %s\n\n", usr_rtn_path);

    /* read routine archive line-by-line */
    while(fgets(line, FM_STR_MAX, archive) != NULL) {

        if(line[strlen(line) - 1] == '\n') line[strlen(line) - 1] = '\0';

        ln++;

        if((ln == 1) || (ln == 2)) {
            printf("%s\n", line);
            continue;
        }

        if(ln == 3) {
            printf("\nRoutines:\n\n");
            next_routine_flag = TRUE;
        }

        if(strcmp(line, "") == 0) {

            next_routine_flag = TRUE;
            
            if(rtn_open == TRUE) {
                fclose(rtn);
                
                rtn_open = FALSE;
            }

            continue;
        }

        if(next_routine_flag == TRUE) {
            
            next_routine_flag = FALSE;

            if((!isalpha(line[0])) && (line[0] != '%')) {

                if(rtn_open == TRUE) {
                    fclose(rtn);
                    rtn_open = FALSE;
                }

                continue; 
            }

            printf(" %s\n", line);

            switch(line[0]) {
                
                case '%':
                    pct_loaded++;

                    snprintf(roufile, PATH_MAX, "%s/%s.m", pct_rtn_path, line);
                    break;

                default:
                    usr_loaded++;

                    snprintf(roufile, PATH_MAX, "%s/%s.m", usr_rtn_path, line);
                    break;
            }

            if((file_exists(roufile) == TRUE) && (rtn_overwrite == FALSE)) {
                fprintf(stdout, "Routine file '%s' already exists. Will not overwrite.\n", roufile);

                return 7;
            }

            if((rtn = fopen(roufile, "w")) == NULL) {
                fprintf(stdout, "Could not open routine file '%s'.\n", roufile);

                return 8;
            }
            else {
                rtn_open = TRUE;
            }

        }
        else {
            if(rtn_open == TRUE) fprintf(rtn, "%s\n", line);
        }

    }

    printf("\n\nLoaded %d user routines and %d percent routines (%d total).\n\n", usr_loaded, pct_loaded, usr_loaded + pct_loaded);

    return 0;

}
